@@ -1,20 +1,32 @@
-# SPDX-License-Identifier: BSD-2-Clause
-
-# Copyright (c) 2025 Phil Thompson <phil@riverbankcomputing.com>
+# Copyright (c) 2022, Riverbank Computing Limited
+# All rights reserved.
+#
+# This copy of SIP is licensed for use under the terms of the SIP License
+# Agreement.  See the file LICENSE for more details.
+#
+# This copy of SIP may also used under the terms of the GNU General Public
+# License v2 or v3 as published by the Free Software Foundation which can be
+# found in the files LICENSE-GPL2 and LICENSE-GPL3 included in this package.
+#
+# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
+# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
+# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
+# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
+# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
+# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
+# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
+# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
+# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
+# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
+# POSSIBILITY OF SUCH DAMAGE.
 
 
-from ...exceptions import UserException
-from ...module import parse_abi_version
-from ...sip_module_configuration import apply_module_option
-
-from ..scoped_name import ScopedName
 from ..specification import (AccessSpecifier, Argument, ArgumentType,
         ArrayArgument, ClassKey, Docstring, DocstringFormat, Extract,
-        FunctionCall, GILUse, IfaceFile, IfaceFileType, KwArgs, License,
-        MappedType, MappedTypeTemplate, MultiInterpreterSupport, Overload,
-        Property, PyQtMethodSpecifier, QualifierType, Signature, Template,
-        ThrowArguments, Value, ValueType, VirtualErrorHandler, WrappedTypedef,
-        WrappedVariable)
+        FunctionCall, IfaceFile, IfaceFileType, KwArgs, License, MappedType,
+        MappedTypeTemplate, Overload, Property, PyQtMethodSpecifier,
+        QualifierType, ScopedName, Signature, Template, ThrowArguments, Value,
+        ValueType, VirtualErrorHandler, WrappedTypedef, WrappedVariable)
 from ..templates import same_template_signature
 from ..utils import cached_name, normalised_scoped_name, search_typedefs
 
@@ -77,7 +89,6 @@ def p_statement(p):
         | license
         | mapped_type
         | mapped_type_template
-        | minimum_abi_version
         | module
         | module_code
         | module_header_code
@@ -85,7 +96,6 @@ def p_statement(p):
         | plugin
         | preinit_code
         | postinit_code
-        | sip_module_configuration
         | timeline
         | type_hint_code
         | unit_code
@@ -96,7 +106,6 @@ def p_statement(p):
 def p_namespace_statement(p):
     """namespace_statement : if_start
         | if_end
-        | namespace_docstring
         | class_decl
         | class_template
         | enum_decl
@@ -191,12 +200,6 @@ def p_composite_module(p):
     if pm.skipping:
         return
 
-    # A composite module must be the first one in the specification.
-    if not pm.in_main_module:
-        pm.parser_error(p, 1, "a %CompositeModule cannot be %Imported")
-
-    pm.spec.is_composite = True
-
     if len(p) == 4:
         name = p[2]
         body = p[3]
@@ -206,7 +209,15 @@ def p_composite_module(p):
 
     module = pm.module_state.module
 
+    if module is not pm.spec.modules[0]:
+        pm.parser_error(p, 1, "a %CompositeModule cannot be %Imported")
+
+    if module.fq_py_name is not None:
+        pm.parser_error(p, 1,
+                "%CompositeModule must appear before any %Module directive")
+
     module.fq_py_name = cached_name(pm.spec, str(name))
+    module.is_composite = True
 
     for directive in body:
         if isinstance(directive, Docstring):
@@ -537,7 +548,7 @@ def p_extract(p):
         order = args.get('order', -1)
         part = p[7]
 
-    pm.spec.extracts.append(Extract(id, order, part.text))
+    pm.spec.extracts.append(Extract(id, order, part))
 
 
 def p_extract_args(p):
@@ -873,7 +884,6 @@ def p_license_arg(p):
 # The mapped type annotations.
 _MAPPED_TYPE_ANNOTATIONS = (
     'AllowNone',
-    'Movable',
     'NoAssignmentOperator',
     'NoCopyCtor',
     'NoDefaultCtor',
@@ -1000,41 +1010,6 @@ def p_mapped_type_function(p):
     pm.validate_function(p, 1, overload)
 
 
-# %MinimumABIVersion ##########################################################
-
-def p_minimum_abi_version(p):
-    """minimum_abi_version : MinimumABIVersion STRING
-        | MinimumABIVersion begin_args '(' version '=' STRING end_args ')'"""
-
-    pm = p.parser.pm
-
-    if pm.skipping:
-        return
-
-    symbol = 2 if len(p) == 3 else 6
-
-    try:
-        min_major_version, min_minor_version = parse_abi_version(p[symbol])
-    except UserException as e:
-        pm.parser_error(p, symbol, e.text)
-        return
-
-    if pm.spec.target_abi is None:
-        pm.spec.target_abi = (min_major_version, min_minor_version)
-    else:
-        major_version, minor_version = pm.spec.target_abi
-
-        if major_version != min_major_version:
-            pm.parser_error(p, symbol,
-                    f"v{min_major_version} is specified but v{major_version} is being targeted")
-        elif min_minor_version is not None:
-            if minor_version is None:
-                pm.spec.target_abi = (min_major_version, min_minor_version)
-            elif min_minor_version is not None and minor_version < min_minor_version:
-                pm.parser_error(p, symbol,
-                        f"v{min_major_version}.{min_minor_version} is the minimum ABI version but v{major_version}.{minor_version} is being targeted")
-
-
 # %ModuleHeaderCode ###########################################################
 
 def p_module_header_code(p):
@@ -1059,15 +1034,22 @@ def p_module(p):
     if pm.skipping:
         return
 
+    module_state = pm.module_state
+    module = module_state.module
+
     # See if this %Module is part of a %CompositeModule.
-    if pm.spec.is_composite:
+    if module.is_composite or module.composite is not None:
         # Historically we %Include modules although conceptually we actually
         # %Import them.  Ensure that the scopes etc. are correct in either
         # case.
         pm.ensure_import()
 
-    module_state = pm.module_state
-    module = module_state.module
+        # The module state may have changed.
+        module_state = pm.module_state
+
+        module_state.module.composite = module if module.is_composite else module.composite
+
+        module = module_state.module
 
     if module.fq_py_name is not None:
         pm.parser_error(p, 1, "%Module has already been specified")
@@ -1094,10 +1076,6 @@ def p_module(p):
     module.default_virtual_error_handler = args.get(
             'default_VirtualErrorHandler')
     module_state.call_super_init = args.get('call_super_init')
-
-    if 'gil_use' in args:
-        module.gil_use = args['gil_use']
-
     module_state.kw_args = args.get('keyword_arguments', KwArgs.NONE)
 
     c_bindings = args.get('language')
@@ -1107,10 +1085,6 @@ def p_module(p):
         elif pm.c_bindings != c_bindings:
             pm.parser_error(p, 1, "cannot mix 'C' and 'C++' modules")
 
-    if 'multi_interpreter_support' in args:
-        module.multi_interpreter_support = args['multi_interpreter_support']
-
-    # Deprecate and remove when Python v3.12 is no longer supported.
     if 'py_ssize_t_clean' in args:
         module.py_ssize_t_clean = args['py_ssize_t_clean']
 
@@ -1141,10 +1115,8 @@ def p_module_arg(p):
     """module_arg : all_raise_py_exception '=' bool_value
         | call_super_init '=' bool_value
         | default_VirtualErrorHandler '=' NAME
-        | gil_use '=' STRING
         | keyword_arguments '=' STRING
         | language '=' STRING
-        | multi_interpreter_support '=' STRING
         | name '=' dotted_name
         | py_ssize_t_clean '=' bool_value
         | use_argument_names '=' bool_value
@@ -1152,21 +1124,8 @@ def p_module_arg(p):
 
     pm = p.parser.pm
 
-    if p[1] == 'gil_use':
-        if p[3] == 'Used':
-            value = GILUse.USED
-        elif p[3] == 'NotUsed':
-            value = GILUse.NOT_USED
-        else:
-            pm.parser_error(p, 3,
-                    "'{0}' is not a valid value for 'gil_use'".format(p[3]))
-
-            # Use any value of the right type.
-            value = GILUse.USED
-
-    elif p[1] == 'keyword_arguments':
+    if p[1] == 'keyword_arguments':
         value = pm.convert_kw_args(p, 3)
-
     elif p[1] == 'language':
         if p[3] == 'C':
             value = True
@@ -1175,22 +1134,6 @@ def p_module_arg(p):
         else:
             pm.parser_error(p, 3, "unsupported language '{0}'".format(p[3]))
             value = None
-
-    elif p[1] == 'multi_interpreter_support':
-        if p[3] == 'NotSupported':
-            value = MultiInterpreterSupport.NOT_SUPPORTED
-        elif p[3] == 'PerInterpreterGILSupported':
-            value = MultiInterpreterSupport.PER_INTERPRETER_GIL_SUPPORTED
-        elif p[3] == 'Supported':
-            value = MultiInterpreterSupport.SUPPORTED
-        else:
-            pm.parser_error(p, 3,
-                    "'{0}' is not a valid value for 'multi_interpreter_support'".format(
-                            p[3]))
-
-            # Use any value of the right type.
-            value = MultiInterpreterSupport.NOT_SUPPORTED
-
     else:
         value = p[3]
 
@@ -1262,7 +1205,7 @@ def p_pickle_code(p):
 # %Platforms ##################################################################
 
 def p_platforms(p):
-    "platforms : Platforms '{' name_list '}'"
+    "platforms : Platforms '{' qualifier_list '}'"
 
     pm = p.parser.pm
 
@@ -1425,30 +1368,10 @@ def p_release_code(p):
     pm.scope.release_code = p[2]
 
 
-# %SipModuleConfiguration #####################################################
-
-def p_sip_module_configuration (p):
-    "sip_module_configuration : SipModuleConfiguration '{' name_list '}'"
-
-    pm = p.parser.pm
-
-    if pm.skipping:
-        return
-
-    spec = pm.spec
-
-    for opt in p[3]:
-        try:
-            spec.sip_module_configuration = apply_module_option(
-                    spec.sip_module_configuration, opt)
-        except UserException as e:
-            pm.parser_error(p, 1, e.text)
-
-
 # %Timeline ###################################################################
 
 def p_timeline(p):
-    "timeline : Timeline '{' name_list '}'"
+    "timeline : Timeline '{' qualifier_list '}'"
 
     pm = p.parser.pm
 
@@ -1488,19 +1411,6 @@ def p_type_code(p):
     pm.scope.type_code.append(p[2])
 
 
-# %TypeDerivedCode ############################################################
-
-def p_type_derived_code(p):
-    "type_derived_code : TypeDerivedCode CODE_BLOCK"
-
-    pm = p.parser.pm
-
-    if pm.skipping:
-        return
-
-    pm.scope.type_derived_code.append(p[2])
-
-
 # %TypeHeaderCode #############################################################
 
 def p_type_header_code(p):
@@ -1694,7 +1604,6 @@ _ONE_WORD_MAP = {
     '...':              ArgumentType.ELLIPSIS,
     'void':             ArgumentType.VOID,
     'bool':             ArgumentType.BOOL,
-    '_Bool':            ArgumentType.BOOL,
     'char':             ArgumentType.STRING,
     'double':           ArgumentType.DOUBLE,
     'float':            ArgumentType.FLOAT,
@@ -1734,7 +1643,6 @@ def p_pod_type(p):
         | float
         | double
         | bool
-        | _Bool
         | char
         | wchar_t
         | void
@@ -1763,9 +1671,6 @@ def p_pod_type(p):
         else:
             p[0] = _UNSIGNED_MAP[p[2]]
     else:
-        if p[1] == 'SIP_SSIZE_T':
-            p.parser.pm.deprecated(p, 1, instead="'Py_ssize_t'")
-
         p[0] = _ONE_WORD_MAP[p[1]]
 
 
@@ -1817,7 +1722,6 @@ _CLASS_ANNOTATIONS = (
     'DelayDtor',
     'Deprecated',
     'ExportDerived',
-    'ExportDerivedLocally',
     'External',
     'FileExtension',
     'Metatype',
@@ -1951,7 +1855,7 @@ def p_superclass_list(p):
 
 
 def p_superclass(p):
-    "superclass : class_access simple_superclass"
+    "superclass : class_access scoped_name"
 
     pm = p.parser.pm
 
@@ -1959,13 +1863,24 @@ def p_superclass(p):
         return
 
     if p[1] != 'public':
-        # In SIP v7 this will be an error.
-        pm.deprecated(p, 1)
-
         p[0] = None
         return
 
-    p[0] = p[2]
+    # This is a hack to allow typedef'ed classes to be used before we have
+    # resolved the typedef definitions.  Unlike elsewhere, we require that the
+    # typedef is defined before being used.
+    ad = Argument(ArgumentType.DEFINED, definition=p[2])
+
+    while ad.type is ArgumentType.DEFINED:
+        ad.type = ArgumentType.NONE
+        search_typedefs(pm.spec, ad.definition, ad)
+
+    if ad.type is not ArgumentType.NONE or len(ad.derefs) != 0 or ad.is_const or ad.is_reference:
+        pm.parser_error(p, 2, "super-class list contains an invalid type")
+
+    # Find the actual class.
+    p[0] = pm.find_class(p, 2, IfaceFileType.CLASS, ad.definition,
+            tmpl_arg=pm.parsing_template)
 
 
 def p_class_access(p):
@@ -1980,46 +1895,6 @@ def p_class_access(p):
     p[0] = p[1]
 
 
-def p_simple_superclass(p):
-    """simple_superclass : scoped_name
-        | scoped_name '<' cpp_types '>'"""
-
-    pm = p.parser.pm
-
-    if pm.skipping:
-        return
-
-    # Handle templates.
-    if len(p) == 5:
-        if pm.parsing_template:
-            # For the moment we just remember enough to identify it later when
-            # we have the values to instantiate it.
-            p[0] = Argument(ArgumentType.TEMPLATE,
-                    definition=Template(p[1], Signature(args=p[3])),
-                    source_location=pm.get_source_location(p, 1))
-        else:
-            pm.parser_error(p, 1,
-                    "super-class templates can only be used in a class template")
-
-        return
-
-    # This is a hack to allow typedef'ed classes to be used before we have
-    # resolved the typedef definitions.  Unlike elsewhere, we require that the
-    # typedef is defined before being used.
-    ad = Argument(ArgumentType.DEFINED, definition=p[1])
-
-    while ad.type is ArgumentType.DEFINED:
-        ad.type = ArgumentType.NONE
-        search_typedefs(pm.spec, ad.definition, ad)
-
-    if ad.type is not ArgumentType.NONE or len(ad.derefs) != 0 or ad.is_const or ad.is_reference:
-        pm.parser_error(p, 1, "super-class list contains an invalid type")
-
-    # Find the actual class.
-    p[0] = pm.find_class(p, 1, IfaceFileType.CLASS, ad.definition,
-            tmpl_arg=pm.parsing_template)
-
-
 def p_opt_class_definition(p):
     """opt_class_definition : '{' opt_class_body '}'
         | empty"""
@@ -2073,24 +1948,12 @@ def p_class_line(p):
         | property
         | release_buffer_code
         | type_code
-        | type_derived_code
         | type_header_code
         | type_hint_code
-        | deprecated_code_directives CODE_BLOCK"""
-
-
-def p_deprecated_code_directives(p):
-    """deprecated_code_directives : BIGetReadBufferCode
-        | BIGetWriteBufferCode
-        | BIGetSegCountCode
-        | BIGetCharBufferCode"""
-
-    pm = p.parser.pm
-
-    if pm.skipping:
-        return
-
-    pm.deprecated(p, 1)
+        | BIGetReadBufferCode CODE_BLOCK
+        | BIGetWriteBufferCode CODE_BLOCK
+        | BIGetSegCountCode CODE_BLOCK
+        | BIGetCharBufferCode CODE_BLOCK"""
 
 
 # The ctor annotations.
@@ -2295,52 +2158,28 @@ _ENUM_MEMBER_ANNOTATIONS = (
 
 
 def p_enum_decl(p):
-    "enum_decl : enum opt_enum_key opt_name opt_enum_base opt_annos '{' opt_enum_body '}' ';'"
+    "enum_decl : enum opt_enum_key opt_name opt_annos '{' opt_enum_body '}' ';'"
 
     pm = p.parser.pm
 
     if pm.skipping:
         return
 
-    pm.check_annotations(p, 5, "enum", _ENUM_ANNOTATIONS)
+    pm.check_annotations(p, 4, "enum", _ENUM_ANNOTATIONS)
 
-    pm.add_enum(p, 1, p[3], p[2], p[4], p[5], p[7])
+    pm.add_enum(p, 1, p[3], p[2], p[4], p[6])
 
 
 def p_opt_enum_key(p):
     """opt_enum_key : class
         | struct
+        | union
         | empty"""
 
-    pm = p.parser.pm
-
-    if pm.skipping:
-        return
-
-    if p[1] is not None:
-        pm.cpp_only(p, 1, "enum keys")
-
     # Return True if the enum is scoped.
     p[0] = p[1] is not None
 
 
-def p_opt_enum_base(p):
-    """opt_enum_base : ':' base_type
-        | empty"""
-
-    pm = p.parser.pm
-
-    if pm.skipping:
-        return
-
-    if len(p) == 3:
-        pm.cpp_only(p, 1, "enum bases")
-
-        p[0] = p[2]
-    else:
-        p[0] = None
-
-
 def p_opt_enum_body(p):
     """opt_enum_body : enum_body
         | empty"""
@@ -2414,7 +2253,7 @@ def p_exception(p):
     pm.check_annotations(p, 4, "exception", _EXCEPTION_ANNOTATIONS)
 
     cpp_name = p[2]
-    py_name = pm.get_py_name(cpp_name.base_name, p[4])
+    py_name = pm.get_py_name(cpp_name, p[4])
     pm.check_attributes(p, 2, py_name)
 
     builtin_base, defined_base = p[3]
@@ -2950,7 +2789,6 @@ def p_opt_exceptions(p):
 
     if p[1] == 'throw':
         p[0] = p[3]
-        p.parser.pm.deprecated(p, 1)
     elif p[1] == 'noexcept':
         p[0] = ThrowArguments()
     else:
@@ -2961,7 +2799,7 @@ def p_opt_exception_list(p):
     """opt_exception_list : exception_list
         | empty"""
 
-    p[0] = ThrowArguments() if p[1] is None else p[1]
+    p[0] = ThrowArguments(arguments=[]) if p[1] is None else p[1]
 
 
 def p_exception_list(p):
@@ -3142,24 +2980,11 @@ def p_opt_namespace_body(p):
     """opt_namespace_body : '{' namespace_body '}'
         | empty"""
 
+
 def p_namespace_body(p):
     """namespace_body : namespace_statement
         | namespace_body namespace_statement"""
 
-def p_namespace_docstring(p):
-    "namespace_docstring : docstring"
-
-    pm = p.parser.pm
-
-    if pm.skipping:
-        return
-
-    if pm.scope.docstring is None:
-        pm.scope.docstring = p[1]
-    else:
-        pm.parser_error(p, 1,
-                "%Docstring has already been defined for this namespace")
-
 
 # C/C++ typedefs. #############################################################
 
@@ -3236,7 +3061,7 @@ _UNION_ANNOTATIONS = (
     'AllowNone',
     'DelayDtor',
     'Deprecated',
-    'ExportDerivedLocally',
+    'ExportDerived',
     'External',
     'FileExtension',
     'Metatype',
@@ -3310,15 +3135,6 @@ def p_variable(p):
     annos_symbol = 3
     body = p[4]
 
-    # We don't currently support /AllowNone/ and /DisallowNone/ for variables.
-    # Generated variable setters support allow None for some (pointer) types
-    # but not others.  For pointer types we explicity disallow None.  This
-    # doesn't affect the generated code but does mean that type hints match
-    # those from previous versions.  This should be changed to properly support
-    # the annotations.
-    if len(type.derefs) != 0:
-        type.disallow_none = True
-
     annotations = p[annos_symbol]
 
     pm.check_annotations(p, annos_symbol, "variable", _VARIABLE_ANNOTATIONS)
@@ -3398,15 +3214,10 @@ def p_annotation(p):
     """annotation : NAME
         | NAME '=' annotation_value"""
 
-    p[0] = {}
-
-    if p.parser.pm.skipping:
-        return
-
     value = None if len(p) == 2 else p[3]
     value = p.parser.pm.validate_annotation(p, 1, value)
 
-    p[0][p[1]] = value
+    p[0] = {p[1]: value}
 
 
 def p_annotation_value(p):
@@ -3557,9 +3368,9 @@ def p_ored_qualifiers(p):
         p[0] = p[1] or pm.evaluate_feature_or_platform(p, 4, inverted=True)
 
 
-def p_name_list(p):
-    """name_list : NAME
-        | name_list NAME"""
+def p_qualifier_list(p):
+    """qualifier_list : NAME
+        | qualifier_list NAME"""
 
     if len(p) == 2:
         value = [p[1]]
