@@ -1,21 +1,39 @@
-# SPDX-License-Identifier: BSD-2-Clause
-
-# Copyright (c) 2025 Phil Thompson <phil@riverbankcomputing.com>
+# Copyright (c) 2022, Riverbank Computing Limited
+# All rights reserved.
+#
+# This copy of SIP is licensed for use under the terms of the SIP License
+# Agreement.  See the file LICENSE for more details.
+#
+# This copy of SIP may also used under the terms of the GNU General Public
+# License v2 or v3 as published by the Free Software Foundation which can be
+# found in the files LICENSE-GPL2 and LICENSE-GPL3 included in this package.
+#
+# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
+# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
+# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
+# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
+# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
+# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
+# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
+# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
+# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
+# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
+# POSSIBILITY OF SUCH DAMAGE.
 
 
 
 from copy import copy
 
 from ..error_log import ErrorLog
+from ..formatters import format_overload_as_cpp
 from ..instantiations import instantiate_type_hints
 from ..python_slots import (is_hash_return_slot, is_int_return_slot,
         is_inplace_number_slot, is_rich_compare_slot, is_ssize_return_slot,
         is_void_return_slot, is_zero_arg_slot)
-from ..scoped_name import ScopedName
-from ..specification import (AccessSpecifier, Argument, ArgumentType,
-        ArrayArgument, ClassKey, Constructor, IfaceFileType, MappedType,
-        Member, PyQtMethodSpecifier, PySlot, Signature, Transfer, ValueType,
-        VirtualHandler, VirtualOverload, VisibleMember, WrappedClass)
+from ..specification import (AccessSpecifier, Argument, ArgumentType, ClassKey,
+        Constructor, IfaceFileType, MappedType, Member, PyQtMethodSpecifier,
+        PySlot, ScopedName, Signature, Transfer, ValueType, VirtualHandler,
+        VirtualOverload, VisibleMember, WrappedClass)
 from ..templates import (encoded_template_name, same_template_signature,
         template_code, template_code_blocks, template_expansions)
 from ..utils import (append_iface_file, argument_as_str, cached_name,
@@ -23,35 +41,32 @@ from ..utils import (append_iface_file, argument_as_str, cached_name,
         same_signature, search_typedefs)
 
 
-def resolve(spec, modules):
+def resolve(spec):
     """ Resolve all types of a parsed specification and create additional views
     so that code can be generated.
     """
 
     error_log = ErrorLog()
-    final_checks = []
 
     # Build the list of all imports for each module.
-    for mod in modules:
+    for mod in spec.modules:
         _set_all_imports(mod, error_log)
 
-        # Set the base name of the module.  This is done for efficiency.
-        mod.py_name = mod.fq_py_name.name.split('.')[-1]
-
     # Set the default meta-type for the main module if it doesn't have one
     # explicitly set.
-    if spec.module.default_metatype is None:
-        for mod in spec.module.all_imports:
+    main_mod = spec.modules[0]
+
+    if main_mod.default_metatype is None:
+        for mod in main_mod.all_imports:
             if mod.default_metatype is None:
                 continue
 
-            if spec.module.default_metatype is None:
-                spec.module.default_metatype = mod.default_metatype
-            elif spec.module.default_metatype is not mod.default_metatype:
+            if main_mod.default_metatype is None:
+                main_mod.default_metatype = mod.default_metatype
+            elif main_mod.default_metatype is not mod.default_metatype:
                 error_log.log(
                         "'{0}' module has imported different default meta-types '{1}' and '{2}'".format(
-                                spec.module.fq_py_name,
-                                spec.module.default_metatype,
+                                main_mod.fq_py_name, main_mod.default_metatype,
                                 mod.default_metatype))
 
     # Check each class has been defined.
@@ -74,14 +89,10 @@ def resolve(spec, modules):
     spec.classes = []
 
     for klass in reversed_classes:
-        # Ignore undefined classes.
-        if klass.iface_file.module is None:
-            continue
-
         _set_mro(spec, klass, error_log)
 
     # Resolve the various types in the modules.
-    _resolve_module(spec, spec.module, error_log, final_checks)
+    _resolve_module(spec, spec.modules[0], error_log)
 
     # Handle default ctors now that the argument types are resolved.
     for klass in spec.classes:
@@ -98,14 +109,18 @@ def resolve(spec, modules):
 
     # Move casts and slots around to their correct classes (if in the same
     # module) or create proxies for them (if cross-module).
-    _move_main_module_casts_slots(spec, error_log)
+    for mod in spec.modules:
+        if _generating_code_for_module(spec, mod):
+            _move_main_module_casts_slots(spec, mod, error_log)
 
     # Automatically generate missing complementary slots.
     for klass in spec.classes:
         _add_complementary_slots(spec, klass)
 
-    for klass in spec.module.proxies:
-        _add_complementary_slots(spec, klass)
+    for mod in spec.modules:
+        if _generating_code_for_module(spec, mod):
+            for klass in mod.proxies:
+                _add_complementary_slots(spec, klass)
 
     # Generate the different class views.
     for klass in spec.classes:
@@ -125,7 +140,7 @@ def resolve(spec, modules):
                 _iface_files_are_used_by_overload(spec, klass.iface_file.used,
                         overload)
 
-    for mod in modules:
+    for mod in spec.modules:
         # Create the list of numbered types sorted by type name.
         _create_sorted_numbered_types(spec, mod, error_log)
 
@@ -138,10 +153,6 @@ def resolve(spec, modules):
 
     # Additional class specific checks.
     for klass in spec.classes:
-        # Ignore undefined classes.
-        if klass.iface_file.module is None:
-            continue
-
         _check_helpers(spec, klass)
         _check_properties(klass, error_log)
 
@@ -151,9 +162,9 @@ def resolve(spec, modules):
 
         # Include the %TypeHeaderCode for exceptions defined in the main
         # module.
-        if spec.target_abi >= (13, 1) or (spec.target_abi >= (12, 9) and spec.target_abi < (13, 0)):
-            if exception_mod is spec.module:
-                append_iface_file(spec.module.used, exception.iface_file)
+        if spec.abi_version >= (13, 1) or (spec.abi_version >= (12, 9) and spec.abi_version < (13, 0)):
+            if exception_mod is spec.modules[0]:
+                append_iface_file(spec.modules[0].used, exception.iface_file)
 
         # Skip those that don't require a Python exception object to be
         # created.
@@ -163,25 +174,30 @@ def resolve(spec, modules):
         if exception.builtin_base_exception is None and exception.defined_base_exception is None:
             continue
 
-        if exception_mod is spec.module or exception.needed:
+        if exception_mod is spec.modules[0] or exception.needed:
             exception.exception_nr = exception_mod.nr_exceptions
             exception_mod.nr_exceptions += 1
 
     # For PyQt6 mark all enum interface files as being used.
     if 'PyQt6' in spec.plugins:
         for enum in spec.enums:
-            if enum.module is spec.module:
-                _enum_iface_file_is_used(enum, spec.module)
+            if enum.module is spec.modules[0]:
+                _enum_iface_file_is_used(enum, spec.modules[0])
 
-    # Perform any final checks.
-    for check in final_checks:
-        check()
+    # Create the name cache as seen by the legacy code.
+    name_cache = spec.name_cache
+    spec.name_cache = []
+
+    for k in sorted(name_cache.keys(), reverse=True):
+        spec.name_cache.extend(sorted(name_cache[k], key=lambda k: k.name))
+
+    _set_string_pool_offsets(spec)
 
     # Raise an exception for any errors.
     error_log.as_exception()
 
 
-def _resolve_module(spec, mod, error_log, final_checks, seen=None):
+def _resolve_module(spec, mod, error_log, seen=None):
     """ Resolve a module and the modules it imports. """
 
     if seen is None:
@@ -194,13 +210,12 @@ def _resolve_module(spec, mod, error_log, final_checks, seen=None):
     # might generate new template-based types and they must be defined in the
     # right module.
     for imported_mod in mod.imports:
-        _resolve_module(spec, imported_mod, error_log, final_checks, seen=seen)
+        _resolve_module(spec, imported_mod, error_log, seen=seen)
 
-    # Resolve typedefs, enums, variables and global functions.
+    # Resolve typedefs, variables and global functions.
     _resolve_typedefs(spec, mod, error_log)
-    _resolve_enums(spec, error_log)
     _resolve_variables(spec, mod, error_log)
-    _resolve_scope_overloads(spec, mod.overloads, error_log, final_checks)
+    _resolve_scope_overloads(spec, mod.overloads, error_log)
 
     # Resolve class ctors, functions and casts.
     for klass in spec.classes:
@@ -211,15 +226,46 @@ def _resolve_module(spec, mod, error_log, final_checks, seen=None):
             _set_needed_exceptions(spec, mod, klass.dtor_throw_args)
 
             _resolve_scope_overloads(spec, klass.overloads, error_log,
-                    final_checks, scope=klass)
+                    scope=klass)
             _transform_casts(spec, klass, error_log)
 
     # Resolve mapped types based on templates.
-    _resolve_mapped_types(spec, mod, error_log, final_checks)
+    _resolve_mapped_types(spec, mod, error_log)
 
     seen.append(mod)
 
 
+def _set_string_pool_offsets(spec):
+    """ Set the offset into the string pool for every used name. """
+
+    offset = 0
+
+    for name in spec.name_cache:
+        if not name.used:
+            continue
+
+        name_len = len(name.name)
+
+        # See if the tail of a previous used name could be used instead.
+        for prev_name in spec.name_cache:
+            prev_name_len = len(prev_name.name)
+
+            if prev_name_len <= name_len:
+                break
+
+            if not prev_name.used or prev_name.is_substring:
+                continue
+
+            if prev_name.name.endswith(name.name):
+                name.is_substring = True
+                name.offset = prev_name.offset + prev_name_len - name_len;
+                break
+
+        if not name.is_substring:
+            name.offset = offset
+            offset += name_len + 1
+
+
 # The map of slots and their compliments.
 _SLOT_MAP = {
     PySlot.LT:  (PySlot.GE, '__ge__'),
@@ -362,21 +408,21 @@ def _set_all_imports(mod, error_log, seen=None):
     seen.remove(mod)
 
 
-def _move_main_module_casts_slots(spec, error_log):
+def _move_main_module_casts_slots(spec, mod, error_log):
     """ Move the casts and slots to the correct place for a main module (ie.
     the one we are generating code for).
     """
 
     for klass in spec.classes:
-        if klass.iface_file.module is spec.module:
-            _move_class_casts(spec, klass, error_log)
+        if klass.iface_file.module is mod:
+            _move_class_casts(spec, mod, klass, error_log)
 
-    for member in spec.module.global_functions:
-        if member.py_slot is not None and member.module is spec.module:
-            _move_global_slot(spec, member, error_log)
+    for member in mod.global_functions:
+        if member.py_slot is not None and member.module is mod:
+            _move_global_slot(spec, mod, member, error_log)
 
 
-def _move_class_casts(spec, klass, error_log):
+def _move_class_casts(spec, mod, klass, error_log):
     """ Move any class casts to its correct class, or publish as a ctor
     extender.
     """
@@ -394,9 +440,9 @@ def _move_class_casts(spec, klass, error_log):
                 cpp_signature=signature, is_cast=True)
 
         # If the destination class is in a different module then use a proxy.
-        if dst_klass.iface_file.module is not spec.module:
-            _iface_file_is_used(spec.module.used, arg)
-            dst_klass = _get_proxy(spec.module, dst_klass)
+        if dst_klass.iface_file.module is not mod:
+            _iface_file_is_used(mod.used, arg)
+            dst_klass = _get_proxy(mod, dst_klass)
             ctor.no_typehint = True
 
         _iface_file_is_used(dst_klass.iface_file.used, arg)
@@ -412,10 +458,10 @@ def _move_class_casts(spec, klass, error_log):
         dst_klass.ctors.append(ctor)
 
 
-def _move_global_slot(spec, global_slot, error_log):
+def _move_global_slot(spec, mod, global_slot, error_log):
     """ If possible, move a global slot to its correct class. """
 
-    for overload in list(spec.module.overloads):
+    for overload in list(mod.overloads):
         if overload.common is not global_slot:
             continue
 
@@ -510,7 +556,7 @@ def _move_global_slot(spec, global_slot, error_log):
                     proxy.members.insert(0, proxy_member)
 
                 # Remove the overload from the list.
-                spec.module.overloads.remove(overload)
+                mod.overloads.remove(overload)
 
                 # Add the overload to the proxy.
                 overload.common = proxy_member
@@ -524,7 +570,7 @@ def _move_global_slot(spec, global_slot, error_log):
             continue
 
         # Remove from the list.
-        spec.module.overloads.remove(overload)
+        mod.overloads.remove(overload)
 
         if arg_enum is not None:
             _enum_iface_file_is_used(arg_enum, arg_module)
@@ -547,12 +593,12 @@ def _move_global_slot(spec, global_slot, error_log):
             # correctly (ie. 'E.M == E.M' works as expected).  However if there
             # is another equality operator defined then it will fail so we have
             # to explicitly inject the comparison.
-            if spec.target_abi < (13, 0) and arg0.type is ArgumentType.ENUM and arg_member.py_slot is PySlot.EQ and not is_second:
+            if spec.abi_version < (13, 0) and arg0.type is ArgumentType.ENUM and arg_member.py_slot is PySlot.EQ and not is_second:
                 inject_equality_slot = True
 
         # Move the overload to the end of the destination list.
         if is_second:
-            overload.is_reflected = True
+            overload.is_reflected
 
         overload.access_specifier = AccessSpecifier.PUBLIC
         overload.common = arg_member
@@ -627,7 +673,7 @@ def _add_auto_overload(spec, auto_klass, auto_overload):
                 overload.is_autogenerated = False
                 klass.overloads.insert(0, overload)
 
-                if klass.iface_file.module is spec.module:
+                if _generating_code_for_module(spec, klass.iface_file.module):
                     member.py_name.used = True
 
                 break
@@ -650,8 +696,8 @@ def _set_mro(spec, klass, error_log, seen=None):
     if klass.scope is not None:
         _set_mro(spec, klass.scope, error_log, seen=seen)
 
-        if klass.scope.deprecated and not klass.deprecated:
-            klass.deprecated = klass.scope.deprecated
+        if klass.scope.deprecated:
+            klass.deprecated = True
 
     if klass.iface_file.type is IfaceFileType.CLASS:
         # The first thing is itself.
@@ -685,11 +731,11 @@ def _set_mro(spec, klass, error_log, seen=None):
                 if superklass_mro not in klass.mro:
                     klass.mro.append(superklass_mro)
 
-                if klass.iface_file.module is spec.module:
+                if _generating_code_for_module(spec, klass.iface_file.module):
                     superklass_mro.iface_file.needed = True
 
-                if superklass_mro.deprecated and not klass.deprecated:
-                    klass.deprecated = superklass_mro.deprecated
+                if superklass_mro.deprecated:
+                    klass.deprecated = True
 
                 # If the super-class is a QObject sub-class then this one is as
                 # well.
@@ -716,19 +762,15 @@ def _set_mro(spec, klass, error_log, seen=None):
         # If the class doesn't have an explicit meta-type then inherit from the
         # module's default.
         if klass.metatype is None and len(klass.superclasses) == 0:
-            # The class may not have been defined.
-            if klass.iface_file.module is not None:
-                klass.metatype = klass.iface_file.module.default_metatype
+            klass.metatype = klass.iface_file.module.default_metatype
 
-        if klass.metatype is not None and klass.iface_file.module is spec.module:
+        if klass.metatype is not None and _generating_code_for_module(spec, klass.iface_file.module):
             klass.metatype.used = True
 
         # If the class doesn't have an explicit super-type then inherit from
         # the module's default.
         if klass.supertype is None and len(klass.superclasses) == 0:
-            # The class may not have been defined.
-            if klass.iface_file.module is not None:
-                klass.supertype = klass.iface_file.module.default_supertype
+            klass.supertype = klass.iface_file.module.default_supertype
 
         if klass.supertype is not None:
             # If the super-type ends with 'sip.wrapper' then assume it is the
@@ -736,7 +778,7 @@ def _set_mro(spec, klass, error_log, seen=None):
             if klass.supertype.name.endswith('sip.wrapper'):
                 klass.supertype = None
 
-        if klass.supertype is not None and klass.iface_file.module is spec.module:
+        if klass.supertype is not None and _generating_code_for_module(spec, klass.iface_file.module):
             klass.supertype.used = True
 
     # Make sure that the module in which a sub-class convertor will be created
@@ -775,7 +817,7 @@ def _resolve_typedefs(spec, mod, error_log):
                     error_log)
 
 
-def _resolve_mapped_types(spec, mod, error_log, final_checks):
+def _resolve_mapped_types(spec, mod, error_log):
     """ Resolve the data types for mapped types based on a template. """
 
     for mapped_type in spec.mapped_types:
@@ -784,7 +826,7 @@ def _resolve_mapped_types(spec, mod, error_log, final_checks):
                 _resolve_mapped_type_types(spec, mapped_type, error_log)
             else:
                 _resolve_scope_overloads(spec, mapped_type.overloads,
-                        error_log, final_checks, scope=mapped_type)
+                        error_log, scope=mapped_type)
 
 
 def _resolve_ctors(spec, klass, error_log):
@@ -817,9 +859,8 @@ def _resolve_ctors(spec, klass, error_log):
                                     klass.iface_file.fq_cpp_name))
                     break
 
-
-        if klass.deprecated and not ctor.deprecated :
-            ctor.deprecated = klass.deprecated   
+        if klass.deprecated:
+            ctor.deprecated = True
 
 
 def _transform_casts(spec, klass, error_log):
@@ -874,9 +915,9 @@ def _add_default_copy_ctor(klass):
     signature = Signature(args=[arg], result=result)
     ctor = Constructor(AccessSpecifier.PUBLIC, py_signature=signature,
             cpp_signature=signature)
-
-    if klass.deprecated and not ctor.deprecated :
-        ctor.deprecated = klass.deprecated
+ 
+    if klass.deprecated:
+        ctor.deprecated = True
 
     if not klass.is_abstract:
         klass.can_create = True
@@ -885,13 +926,12 @@ def _add_default_copy_ctor(klass):
     klass.ctors.append(ctor)
 
 
-def _resolve_scope_overloads(spec, overloads, error_log, final_checks,
-        scope=None):
+def _resolve_scope_overloads(spec, overloads, error_log, scope=None):
     """ Resolve the data types for a scope's overloads. """
 
     for overload in overloads:
         _resolve_func_types(spec, overload.common.module, scope, overload,
-                error_log, final_checks)
+                error_log)
 
         # Now check that the Python signature doesn't conflict with an earlier
         # one.  If there is %MethodCode then assume that it will handle any
@@ -921,53 +961,13 @@ def _resolve_scope_overloads(spec, overloads, error_log, final_checks,
                     break
 
         if isinstance(scope, WrappedClass):
-            if scope.deprecated and not overload.deprecated :
-                overload.deprecated = scope.deprecated
+            if scope.deprecated:
+                overload.deprecated = True
 
             if overload.is_abstract:
                 scope.is_abstract = True
 
 
-# The supported enum base types.  Note that we use the STRING types and the
-# BYTE types because there is no opportunity to apply the /PyInt/ annotation
-# and there can be no confusion about context.  We still need the BYTE types
-# because /PyInt/ could have been specified in a typedef.
-_ENUM_BASE_TYPES = (
-    ArgumentType.STRING,
-    ArgumentType.SSTRING,
-    ArgumentType.USTRING,
-    ArgumentType.BYTE,
-    ArgumentType.SBYTE,
-    ArgumentType.UBYTE,
-    ArgumentType.SHORT,
-    ArgumentType.USHORT,
-    ArgumentType.INT,
-    ArgumentType.UINT,
-    ArgumentType.BOOL,
-)
-
-def _resolve_enums(spec, error_log):
-    """ Resolve the base types for all the enums. """
-
-    for enum in spec.enums:
-        base_type = enum.enum_base_type
-
-        if base_type is None:
-            continue
-
-        _resolve_type(spec, enum.module, enum.scope, base_type, error_log)
-
-        # The current ABI implementations only support enums no larger than an
-        # int.
-        if base_type.type not in _ENUM_BASE_TYPES or len(base_type.derefs) != 0:
-            error_log.log(f"unsupported enum base type",
-                    source_location=base_type.source_location)
-
-        # The default is int.
-        if base_type.type is ArgumentType.INT:
-            enum.enum_base_type = None
-
-
 def _resolve_variables(spec, mod, error_log):
     """ Resolve the data types for the variables of a module. """
 
@@ -1002,7 +1002,7 @@ def _get_visible_py_members(spec, klass):
                         if overload.is_abstract:
                             klass.is_abstract = True
 
-                        if klass.iface_file.module is spec.module and (klass is mro_klass or (overload.access_specifier is AccessSpecifier.PROTECTED and klass.has_shadow)):
+                        if _generating_code_for_module(spec, klass.iface_file.module) and (klass is mro_klass or (overload.access_specifier is AccessSpecifier.PROTECTED and klass.has_shadow)):
                             need_types = True
                             member.py_name.used = True
 
@@ -1069,7 +1069,7 @@ def _add_virtual_overload(spec, overload, klass, error_log):
 
     # If this class is defined in the main module then make sure the virtuals
     # have a handler.
-    if klass.iface_file.module is spec.module:
+    if _generating_code_for_module(spec, klass.iface_file.module):
         virtual_handler = _get_virtual_handler(spec, overload, klass,
                 error_log)
 
@@ -1078,7 +1078,7 @@ def _add_virtual_overload(spec, overload, klass, error_log):
 
         # Make sure we have the interface files and type definitions for the
         # virtual handler.
-        _iface_files_are_used_by_overload(spec, spec.module.used, overload,
+        _iface_files_are_used_by_overload(spec, spec.modules[0].used, overload,
                 need_types=True)
     else:
         virtual_handler = None
@@ -1261,7 +1261,7 @@ def _resolve_ctor_types(spec, scope, ctor, error_log):
         _scope_default_value(spec, scope, arg)
 
 
-def _resolve_func_types(spec, mod, scope, overload, error_log, final_checks):
+def _resolve_func_types(spec, mod, scope, overload, error_log):
     """ Resolve the types of a function. """
 
     # Handle any exceptions.
@@ -1282,14 +1282,13 @@ def _resolve_func_types(spec, mod, scope, overload, error_log, final_checks):
             _resolve_type(spec, mod, scope, arg, error_log, allow_defined=True)
  
     # Handle the Python signature.
-    _resolve_py_signature_types(spec, mod, scope, overload, error_log,
-            final_checks)
+    _resolve_py_signature_types(spec, mod, scope, overload, error_log)
 
     result = overload.py_signature.result
 
     # These slots must return Py_ssize_t.
     if is_ssize_return_slot(overload.common.py_slot):
-        if spec.target_abi >= (13, 0):
+        if spec.abi_version >= (13, 0):
             required_types = (ArgumentType.SSIZE, )
         else:
             required_types = (ArgumentType.SSIZE, ArgumentType.INT)
@@ -1312,7 +1311,7 @@ def _resolve_func_types(spec, mod, scope, overload, error_log, final_checks):
 
     # These slots must return Py_hash_t.
     if is_hash_return_slot(overload.common.py_slot):
-        if spec.target_abi >= (13, 0):
+        if spec.abi_version >= (13, 0):
             required_type = ArgumentType.HASH
             required_type_name = 'Py_hash_t'
         else:
@@ -1325,8 +1324,7 @@ def _resolve_func_types(spec, mod, scope, overload, error_log, final_checks):
                     scope=scope)
 
 
-def _resolve_py_signature_types(spec, mod, scope, overload, error_log,
-        final_checks):
+def _resolve_py_signature_types(spec, mod, scope, overload, error_log):
     """ Resolve the types of a Python signature. """
 
     result = overload.py_signature.result
@@ -1376,48 +1374,10 @@ def _resolve_py_signature_types(spec, mod, scope, overload, error_log,
                             arg_nr + 1),
                     overload, scope=scope)
 
-        # Check that the argument support /Array/.
-        if arg.type in (ArgumentType.CLASS, ArgumentType.MAPPED) and arg.array is ArrayArgument.ARRAY:
-            # This is delayed because the class must be complete.  Be careful
-            # not to pass any non-local changing variables.
-            local_arg_nr = arg_nr
-            final_checks.append(
-                    lambda: _check_array_support(overload, local_arg_nr, scope,
-                            error_log))
-
-        # If the type of the argument is a /Movable/ mapped type then add
-        # /Transfer/.
-        if arg.type is ArgumentType.MAPPED and arg.definition.movable:
-            arg.transfer = Transfer.TRANSFER
-
         if scope is not None:
             _scope_default_value(spec, scope, arg)
 
 
-def _check_array_support(overload, arg_nr, scope, error_log):
-    """ Check that an argument supports /Array/. """
-
-    arg = overload.py_signature.args[arg_nr]
-
-    if len(arg.derefs) != 1 or not arg.is_in or arg.is_reference:
-        _log_overload_error(error_log,
-                "argument {0} is a mapped type or class with /Array/ and so must be a pointer".format(
-                        arg_nr + 1),
-                overload, scope=scope)
-
-    if arg.type is ArgumentType.MAPPED and arg.definition.no_release:
-        _log_overload_error(error_log,
-                "argument {0} is a mapped type that does not support /Array/ because /NoRelease/ has been specified for it".format(
-                        arg_nr + 1),
-                overload, scope=scope)
-
-    if arg.type is ArgumentType.CLASS and not arg.definition.needs_array_helper:
-        _log_overload_error(error_log,
-                "argument {0} is a class that does not support /Array/".format(
-                        arg_nr + 1),
-                overload, scope=scope)
-
-
 # Various type classifications.
 _CLASS_TYPES = (ArgumentType.MAPPED, ArgumentType.CLASS)
 _SIMPLE_TYPES = (ArgumentType.CFLOAT, ArgumentType.FLOAT, ArgumentType.CDOUBLE,
@@ -1801,7 +1761,7 @@ def _resolve_type(spec, mod, scope, type, error_log, allow_defined=False):
 
     # If we are in the main module then mark any generated types as being
     # needed.
-    if mod is spec.module:
+    if _generating_code_for_module(spec, mod):
         _set_needed_type(type)
 
 
@@ -1817,7 +1777,7 @@ def _set_needed_type(arg):
 def _set_needed_exceptions(spec, mod, throw_args):
     """ Specify that a set of thrown arguments are needed. """
 
-    if mod is spec.module and throw_args is not None and throw_args.arguments is not None:
+    if _generating_code_for_module(spec, mod) and throw_args is not None and throw_args.arguments is not None:
         for exception in throw_args.arguments:
             _set_needs_exception(exception)
 
@@ -1872,13 +1832,12 @@ def _instantiate_mapped_type_template(spec, mod, mapped_type_template, type,
     mapped_type.type.is_reference = False
     mapped_type.cpp_name = cached_name(spec, argument_as_str(mapped_type.type))
 
-    if mod is spec.module:
+    if _generating_code_for_module(spec, mod):
         mapped_type.cpp_name.used = True
 
     proto_mapped_type = mapped_type_template.mapped_type
 
     mapped_type.handles_none = proto_mapped_type.handles_none
-    mapped_type.movable = proto_mapped_type.movable
     mapped_type.needs_user_state = proto_mapped_type.needs_user_state
     mapped_type.no_assignment_operator = proto_mapped_type.no_assignment_operator
     mapped_type.no_copy_ctor = proto_mapped_type.no_copy_ctor
@@ -1934,7 +1893,7 @@ def _search_scope(spec, scope, scoped_name, type):
 
 
 def _name_lookup(spec, mod, scoped_name, type):
-    """ Look up a name and resolve the corresponding type. """
+    """ Look up a name and resole the corresponding type. """
 
     _search_mapped_types(spec, mod, type, scoped_name)
     if type.type is not ArgumentType.NONE:
@@ -2082,16 +2041,17 @@ def _iface_files_are_used_by_overload(spec, used, overload, need_types=False):
                 need_types=need_types)
 
     # Don't bother with %TypeHeaderCode from %Exception for later ABI versions.
-    if spec.target_abi >= (13, 1) or (spec.target_abi >= (12, 9) and spec.target_abi < (13, 0)):
+    if spec.abi_version >= (13, 1) or (spec.abi_version >= (12, 9) and spec.abi_version < (13, 0)):
         return
 
-    throw_args = overload.throw_args
-    if throw_args is not None and throw_args.arguments is not None:
-        for exception in throw_args.arguments:
-            append_iface_file(used, exception.iface_file)
+    if overload.throw_args is not None:
+        for throw_arg in overload.throw_args:
+            if throw_arg is not None:
+                for exception in throw_arg.arguments:
+                    append_iface_file(used, exception.iface_file)
 
-            if need_types:
-                _set_needs_exception(exception)
+                    if need_types:
+                        _set_needs_exception(exception)
 
 
 def _iface_file_is_used(used, arg, need_types=False):
@@ -2152,7 +2112,7 @@ def _create_sorted_numbered_types(spec, mod, error_log):
         if klass.iface_file.module is not mod:
             continue
 
-        if mod is spec.module or klass.iface_file.needed:
+        if _generating_code_for_module(spec, mod) or klass.iface_file.needed:
             if not klass.is_hidden_namespace:
                 mod.needed_types.append(Argument(ArgumentType.CLASS,
                         definition=klass, name=klass.iface_file.cpp_name))
@@ -2161,7 +2121,7 @@ def _create_sorted_numbered_types(spec, mod, error_log):
         if mapped_type.iface_file.module is not mod:
             continue
 
-        if mod is spec.module or mapped_type.iface_file.needed:
+        if _generating_code_for_module(spec, mod) or mapped_type.iface_file.needed:
             mod.needed_types.append(Argument(ArgumentType.MAPPED,
                     definition=mapped_type, name=mapped_type.cpp_name))
 
@@ -2172,7 +2132,7 @@ def _create_sorted_numbered_types(spec, mod, error_log):
         if enum.fq_cpp_name is None:
             continue
 
-        if mod is spec.module or enum.needed:
+        if _generating_code_for_module(spec, mod) or enum.needed:
             mod.needed_types.append(Argument(ArgumentType.ENUM,
                     definition=enum, name=enum.cached_fq_cpp_name))
 
@@ -2202,6 +2162,14 @@ def _create_sorted_numbered_types(spec, mod, error_log):
         needed_type_nr += 1
 
 
+def _generating_code_for_module(spec, mod):
+    """ Return True if we are generating code for a module, ie. we are the main
+    module.
+    """
+
+    return spec.modules[0] is mod
+
+
 def _check_properties(klass, error_log):
     """ Check that any properties are valid. """
 
@@ -2220,10 +2188,12 @@ def _check_properties(klass, error_log):
 def _log_overload_error(error_log, text, overload, scope=None):
     """ Log an error about an overload. """
 
-    if scope is None:
-        fq_cpp_name = overload.cpp_name
+    if scope is not None:
+        scope_s = str(scope.iface_file.fq_cpp_name) + '::'
     else:
-        fq_cpp_name = f'{scope.iface_file.fq_cpp_name}::{overload.cpp_name}'
+        scope_s = ''
 
-    error_log.log(f"'{fq_cpp_name}' {text}",
+    error_log.log(
+            "'{0}{1}' {2}".format(scope_s, format_overload_as_cpp(overload),
+                    text),
             source_location=overload.source_location)
